import os
from platform import python_version, platform

from rich.console import Console

from .terminal import Notify, Style
from ..api import Api, SORT_CRITERION, TIMEFRAME, TIME_FORMAT
from ..meta import about, version

__all__ = [
    "api",
    "console",
    "OUTPUT_PARENT_DIR",
    "notify",
    "style",
    "SORT_CRITERION",
    "TIMEFRAME",
    "TIME_FORMAT",
]

api = Api(
    headers={
        "User-Agent": f"{about.name.replace(' ', '-')}/{version.release} "
        f"(Python {python_version} on {platform}; +{about.documentation})"
    },
)

console = Console(log_time=False)

notify = Notify(console=console)
style = Style

OUTPUT_PARENT_DIR: str = os.path.expanduser(os.path.join("~", "knewkarma"))
ML_MODELS_DIR: str = os.path.join(OUTPUT_PARENT_DIR, "ml_models")



import os
from datetime import datetime
from typing import Union, List, Optional

from rich.box import DOUBLE
from rich.console import ConsoleRenderable, RichCast
from rich.panel import Panel

from .shared import console, notify

__all__ = [
    "filename_timestamp",
    "make_panel",
    "pathfinder",
]


def make_panel(
    title: str,
    content: Union[ConsoleRenderable, RichCast, str],
    subtitle: Optional[str] = None,
):
    """
    Makes a rich Panel for whatever data is needed to be placed in it.

    :param title: Panel title.
    :type title: str
    :param content: Content of the panel.
    :type content: Union[rich.ConsoleRenderable, rich.RichCast, str]
    :param subtitle: Panel subtitle.
    :type subtitle: str
    """
    from .terminal import Style

    console.print(
        Panel(
            renderable=content,
            title=f"{Style.bold}{title}{Style.reset}",
            subtitle=(subtitle if subtitle else None),
            box=DOUBLE,
            style=f"{Style.white.strip('[,]')} on black",
        )
    )


def filename_timestamp() -> str:
    """
    Generates a timestamp string suitable for file naming, based on the current date and time.
    The format of the timestamp is adapted based on the operating system.

    :return: The formatted timestamp as a string. The format is "%d-%B-%Y-%I-%M-%S%p" for Windows
             and "%d-%B-%Y-%I:%M:%S%p" for non-Windows systems.
    :rtype: str

    Example
    -------
    - Windows: "20-July-1969-08-17-45PM"
    - Non-Windows: "20-July-1969-08:17:45PM"
    """
    now = datetime.now()
    return (
        now.strftime("%d-%B-%Y-%I-%M-%S%p")
        if os.name == "nt"
        else now.strftime("%d-%B-%Y-%I:%M:%S%p")
    )


def pathfinder(directories: Union[List[str], str]):
    """
    Creates directories for exported data (`exported`).

    :param directories: A list of directories or a directory name to create
    :type directories: Union[List[str], str]
    :raise Exception: If any was encountered.
    """

    try:
        if isinstance(directories, List) and all(
            isinstance(directory, str) for directory in directories
        ):
            for directory in directories:
                os.makedirs(name=directory, exist_ok=True)
        elif isinstance(directories, str):
            os.makedirs(name=directories, exist_ok=True)
    except Exception as unexpected_error:
        notify.exception(unexpected_error)


# -------------------------------- END ----------------------------------------- #



from .version import full as full_version

__all__ = [
    "author",
    "copyright",
    "description",
    "documentation",
    "name",
    "package",
    "sponsor",
    "summary",
]

package: str = "knewkarma"
name: str = "Knew Karma"
author: tuple[str, str, str] = (
    "Richard Mwewa",
    "rly0nheart",
    "https://gravatar.com/rly0nheart",
)
sponsor: str = f"https://opencollective.com/{name}"
copyright: str = f"{name} {full_version} © {author[0]}. All rights reserved."
documentation: str = "https://knewkarma.readthedocs.io"
summary: str = f"A Reddit data analysis toolkit — by {author[0]} <{author[2]}>"
description: str = f"""
{name} (/nuː ‘kɑːrmə/) is a Reddit data analysis toolkit designed to provide an extensive range of
functionalities for exploring and analysing Reddit data. It includes a Command-Line Interface (CLI), and an
Application Programming Interface (API) to enable an easy integration in other Python Projects.



📚 Refer to the documentation <{documentation}> for usage examples and integration guide.

❤️  Become a sponsor: <{sponsor}>
"""

# -------------------------------- END ----------------------------------------- #



import re
from collections import Counter
from types import SimpleNamespace
from typing import Literal, Union, Optional, List

from aiohttp import ClientSession
from karmakaze import Parse
from rich.status import Status

from .tools.data import plot_bar_chart, visualisation_dependency_installed
from .tools.shared import api, SORT_CRITERION, TIMEFRAME, TIME_FORMAT

__all__ = [
    "Post",
    "Posts",
    "Search",
    "Subreddit",
    "Subreddits",
    "User",
    "Users",
]


class Comment:
    """Represents a Reddit comment and provides methods for interacting with it."""

    pass


class Post:
    """Represents a Reddit post and provides methods for fetching post data and comments."""

    def __init__(self, id: str, subreddit: str, time_format: TIME_FORMAT = "locale"):
        """
        Initialises a `Post` instance to fetch the post's data and comments.

        :param id: The ID of the Reddit post to retrieve.
        :type id: str
        :param subreddit: The subreddit where the post was created.
        :type subreddit: str
        :param time_format: Format for displaying time, either 'concise' or 'locale'. Defaults to 'locale'.
        :type time_format: Literal["concise", "locale"]
        """

        self._id = id
        self._subreddit = subreddit
        self._parse = Parse(time_format=time_format)

    async def data(
        self, session: ClientSession, status: Optional[Status] = None
    ) -> SimpleNamespace:
        """
        Asynchronously retrieves data for a Reddit post, excluding comments.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A `SimpleNamespace` object containing parsed post data.
        :rtype: SimpleNamespace
        """

        post_data = await api.get_entity(
            session=session,
            kind="post",
            status=status,
            id=self._id,
            subreddit=self._subreddit,
        )

        parsed_post = self._parse.post(post_data)

        return parsed_post if post_data else SimpleNamespace

    async def comments(
        self,
        session: ClientSession,
        limit: int,
        sort: SORT_CRITERION = "all",
        status: Optional[Status] = None,
    ) -> List[SimpleNamespace]:
        """
        Asynchronously retrieves comments for a Reddit post.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param limit: Maximum number of comments to retrieve.
        :type limit: int
        :param sort: The sorting criterion for the comments. Defaults to "all".
        :type sort: SORT_CRITERION, optional
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing parsed comment data.
        :rtype: List[SimpleNamespace]
        """

        comments_data = await api.get_posts_or_comments(
            session=session,
            kind="comments_from_a_post",
            status=status,
            id=self._id,
            subreddit=self._subreddit,
            limit=limit,
            sort=sort,
        )

        parsed_comments = self._parse.comments(comments_data)

        return parsed_comments if comments_data else [SimpleNamespace]


class Posts:
    """Represents Reddit posts and provides methods for retrieving posts from various sources."""

    def __init__(self, time_format: TIME_FORMAT = "locale"):
        """
        Initialises a `Posts` instance for retrieving posts such as 'best', 'controversial',
        'front-page', 'new', 'popular', and 'rising'.

        :param time_format: Format for displaying time, either 'concise' or 'locale'. Defaults to 'locale'.
        :type time_format: Literal["concise", "locale"]
        """
        self._parse = Parse(time_format=time_format)

    async def best(
        self,
        session: ClientSession,
        limit: int,
        timeframe: TIMEFRAME = "all",
        status: Optional[Status] = None,
    ) -> List[SimpleNamespace]:
        """
        Asynchronously retrieves the best posts.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param limit: Maximum number of posts to retrieve.
        :type limit: int
        :param timeframe: The timeframe from which to retrieve posts. Defaults to "all".
        :type timeframe: TIMEFRAME, optional
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing parsed post data.
        :rtype: List[SimpleNamespace]
        """

        best_posts = await api.get_posts_or_comments(
            session=session,
            kind="best",
            status=status,
            timeframe=timeframe,
            limit=limit,
        )

        parsed_posts = self._parse.posts(best_posts)

        return parsed_posts if best_posts else [SimpleNamespace]

    async def controversial(
        self,
        session: ClientSession,
        limit: int,
        timeframe: TIMEFRAME = "all",
        status: Optional[Status] = None,
    ) -> List[SimpleNamespace]:
        """
        Asynchronously retrieves the controversial posts.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param limit: Maximum number of posts to retrieve.
        :type limit: int
        :param timeframe: The timeframe from which to retrieve posts. Defaults to "all".
        :type timeframe: TIMEFRAME, optional
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing parsed post data.
        :rtype: List[SimpleNamespace]
        """

        controversial_posts = await api.get_posts_or_comments(
            session=session,
            kind="controversial",
            status=status,
            timeframe=timeframe,
            limit=limit,
        )

        parsed_posts = self._parse.posts(controversial_posts)

        return parsed_posts if controversial_posts else [SimpleNamespace]

    async def front_page(
        self,
        session: ClientSession,
        limit: int,
        timeframe: TIMEFRAME = "all",
        sort: SORT_CRITERION = "all",
        status: Optional[Status] = None,
    ) -> List[SimpleNamespace]:
        """
        Asynchronously retrieves the front-page posts.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param limit: Maximum number of posts to retrieve.
        :type limit: int
        :param timeframe: The timeframe from which to retrieve posts. Defaults to "all".
        :type timeframe: TIMEFRAME, optional
        :param sort: Sorting criterion for posts. Defaults to "all".
        :type sort: SORT_CRITERION, optional
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing parsed post data.
        :rtype: List[SimpleNamespace]
        """

        front_page_posts = await api.get_posts_or_comments(
            session=session,
            kind="front_page",
            status=status,
            limit=limit,
            timeframe=timeframe,
            sort=sort,
        )

        parsed_posts = self._parse.posts(front_page_posts)

        return parsed_posts if front_page_posts else [SimpleNamespace]

    async def new(
        self,
        session: ClientSession,
        limit: int,
        timeframe: TIMEFRAME = "all",
        sort: SORT_CRITERION = "all",
        status: Optional[Status] = None,
    ) -> List[SimpleNamespace]:
        """
        Asynchronously retrieves the new posts.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param limit: Maximum number of posts to retrieve.
        :type limit: int
        :param timeframe: The timeframe from which to retrieve posts. Defaults to "all".
        :type timeframe: TIMEFRAME, optional
        :param sort: Sorting criterion for posts. Defaults to "all".
        :type sort: SORT_CRITERION, optional
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing parsed post data.
        :rtype: List[SimpleNamespace]
        """

        new_posts = await api.get_posts_or_comments(
            session=session,
            kind="new",
            status=status,
            limit=limit,
            timeframe=timeframe,
            sort=sort,
        )

        parsed_posts = self._parse.posts(new_posts)

        return parsed_posts if new_posts else [SimpleNamespace]

    async def popular(
        self,
        session: ClientSession,
        limit: int,
        timeframe: TIMEFRAME = "all",
        status: Optional[Status] = None,
    ) -> List[SimpleNamespace]:
        """
        Asynchronously retrieves the popular posts.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param limit: Maximum number of posts to retrieve.
        :type limit: int
        :param timeframe: The timeframe from which to retrieve posts. Defaults to "all".
        :type timeframe: TIMEFRAME, optional
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing parsed post data.
        :rtype: List[SimpleNamespace]
        """
        popular_posts = await api.get_posts_or_comments(
            session=session,
            kind="popular",
            status=status,
            timeframe=timeframe,
            limit=limit,
        )

        parsed_posts = self._parse.posts(popular_posts)

        return parsed_posts if popular_posts else [SimpleNamespace]

    async def rising(
        self,
        session: ClientSession,
        limit: int,
        timeframe: TIMEFRAME = "all",
        status: Optional[Status] = None,
    ) -> List[SimpleNamespace]:
        """
        Asynchronously retrieves the rising posts.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param limit: Maximum number of posts to retrieve.
        :type limit: int
        :param timeframe: The timeframe from which to retrieve posts. Defaults to "all".
        :type timeframe: TIMEFRAME, optional
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing parsed post data.
        :rtype: List[SimpleNamespace]
        """

        rising_posts = await api.get_posts_or_comments(
            session=session,
            kind="rising",
            status=status,
            timeframe=timeframe,
            limit=limit,
        )

        parsed_posts = self._parse.posts(rising_posts)

        return parsed_posts if rising_posts else [SimpleNamespace]


class Search:
    """
    Represents the Reddit search functionality and provides methods for retrieving search results
    from different entities.
    """

    def __init__(self, query: str, time_format: TIME_FORMAT = "locale"):
        """
        Initialises the `Search` instance for searching posts, subreddits, and users.

        :param query: The search query string.
        :type query: str
        :param time_format: Format for displaying time, either 'concise' or 'locale'. Defaults to 'locale'.
        :type time_format: Literal["concise", "locale"]
        """

        self._query = query
        self._parse = Parse(time_format=time_format)

    async def posts(
        self,
        session: ClientSession,
        limit: int,
        sort: SORT_CRITERION = "all",
        status: Optional[Status] = None,
    ) -> List[SimpleNamespace]:
        """
        Asynchronously retrieves posts that match with the specified query.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param limit: Maximum number of posts to retrieve.
        :type limit: int
        :param sort: Sorting criterion for posts. Defaults to "all".
        :type sort: SORT_CRITERION, optional
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing parsed post data.
        :rtype: List[SimpleNamespace]
        """

        search_results = await api.search_entities(
            session=session,
            kind="posts",
            status=status,
            query=self._query,
            sort=sort,
            limit=limit,
        )

        parsed_posts = self._parse.posts(search_results)

        return parsed_posts if search_results else [SimpleNamespace]

    async def subreddits(
        self,
        session: ClientSession,
        limit: int,
        sort: SORT_CRITERION = "all",
        status: Optional[Status] = None,
    ) -> List[SimpleNamespace]:
        """
        Asynchronously retrieves subreddits that match with the specified query.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param limit: Maximum number of subreddits to retrieve.
        :type limit: int
        :param sort: Sorting criterion for subreddits. Defaults to "all".
        :type sort: SORT_CRITERION, optional
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing parsed subreddit data.
        :rtype: List[SimpleNamespace]
        """

        search_results = await api.search_entities(
            session=session,
            kind="subreddits",
            status=status,
            query=self._query,
            sort=sort,
            limit=limit,
        )

        parsed_subreddits = self._parse.subreddits(search_results)

        return parsed_subreddits if search_results else [SimpleNamespace]

    async def users(
        self,
        session: ClientSession,
        limit: int,
        sort: SORT_CRITERION = "all",
        status: Optional[Status] = None,
    ) -> List[SimpleNamespace]:
        """
        Asynchronously retrieves users that match with the specified query.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param limit: Maximum number of users to retrieve.
        :type limit: int
        :param sort: Sorting criterion for users. Defaults to "all".
        :type sort: SORT_CRITERION, optional
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing parsed user data.
        :rtype: List[SimpleNamespace]
        """

        search_results = await api.search_entities(
            session=session,
            kind="users",
            status=status,
            query=self._query,
            sort=sort,
            limit=limit,
        )

        parsed_users = self._parse.users(search_results)

        return parsed_users if search_results else [SimpleNamespace]


class Subreddit:
    """Represents a Reddit community (subreddit) and provides methods for retrieving its data."""

    def __init__(self, name: str, time_format: TIME_FORMAT = "locale"):
        """
        Initialises a `Subreddit` instance to get a subreddit's profile, wiki pages, and posts data,
        and to search for posts containing a specified query.

        :param name: The name of the subreddit to retrieve data from.
        :type name: str
        :param time_format: Format for displaying time, either 'concise' or 'locale'. Defaults to 'locale'.
        :type time_format: Literal["concise", "locale"]
        """

        self._name = name
        self._time_format = (
            time_format  # This will also be accessed in the comments method
        )
        self._parse = Parse(time_format=self._time_format)

    async def comments(
        self,
        session: ClientSession,
        posts_limit: int,
        comments_per_post: int,
        sort: SORT_CRITERION = "all",
        timeframe: TIMEFRAME = "all",
        status: Optional[Status] = None,
    ) -> List[SimpleNamespace]:
        """
        Asynchronously retrieves comments from a subreddit.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param posts_limit: Maximum number of posts to retrieve comments from.
        :type posts_limit: int
        :param comments_per_post: Maximum number of comments to retrieve per post.
        :type comments_per_post: int
        :param sort: Sorting criterion for the posts and comments. Defaults to "all".
        :type sort: SORT_CRITERION, optional
        :param timeframe: The timeframe from which to retrieve posts and comments. Defaults to "all".
        :type timeframe: TIMEFRAME, optional
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing parsed comment data.
        :rtype: List[SimpleNamespace]
        """

        posts = await self.posts(
            session=session,
            status=status,
            limit=posts_limit,
            sort=sort,
            timeframe=timeframe,
        )
        all_comments: List = []
        for post in posts:
            post = Post(
                id=post.get("id"),
                subreddit=post.get("subreddit"),
                time_format=self._time_format,
            )
            post_comments = await post.comments(
                session=session, limit=comments_per_post, sort=sort, status=status
            )

            all_comments.extend(post_comments)

        return all_comments

    async def posts(
        self,
        session: ClientSession,
        limit: int,
        sort: SORT_CRITERION = "all",
        timeframe: TIMEFRAME = "all",
        status: Optional[Status] = None,
    ) -> List[SimpleNamespace]:
        """
        Asynchronously retrieves posts from a subreddit.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param limit: Maximum number of posts to retrieve.
        :type limit: int
        :param sort: Sorting criterion for the posts. Defaults to "all".
        :type sort: SORT_CRITERION, optional
        :param timeframe: The timeframe from which to retrieve posts. Defaults to "all".
        :type timeframe: TIMEFRAME, optional
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing parsed post data.
        :rtype: List[SimpleNamespace]
        """

        subreddit_posts = await api.get_posts_or_comments(
            session=session,
            kind="posts_from_a_subreddit",
            subreddit=self._name,
            status=status,
            limit=limit,
            sort=sort,
            timeframe=timeframe,
        )

        parsed_posts = self._parse.posts(subreddit_posts)

        return parsed_posts if subreddit_posts else [SimpleNamespace]

    async def profile(
        self, session: ClientSession, status: Optional[Status] = None
    ) -> SimpleNamespace:
        """
        Asynchronously retrieves a subreddit's profile data.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A `SimpleNamespace` object containing the parsed subreddit profile data.
        :rtype: SimpleNamespace
        """

        subreddit_profile = await api.get_entity(
            session=session,
            status=status,
            kind="subreddit",
            subreddit=self._name,
        )

        parsed_profile = self._parse.subreddit(subreddit_profile)

        return parsed_profile if subreddit_profile else SimpleNamespace

    async def search_comments(
        self,
        session: ClientSession,
        query: str,
        posts_limit: int,
        comments_per_post: int,
        sort: SORT_CRITERION = "all",
        timeframe: TIMEFRAME = "all",
        status: Optional[Status] = None,
    ) -> List[SimpleNamespace]:
        """
        Asynchronously retrieves comments that contain the specified query from a subreddit.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param query: Search query.
        :type query: str
        :param posts_limit: Maximum number of posts to retrieve comments from.
        :type posts_limit: int
        :param comments_per_post: Maximum number of comments to retrieve for each post.
        :type comments_per_post: int
        :param sort: Sorting criterion for the comments. Defaults to "all".
        :type sort: SORT_CRITERION, optional
        :param timeframe: The timeframe from which to retrieve comments. Defaults to "all".
        :type timeframe: TIMEFRAME, optional
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing comment data.
        :rtype: List[SimpleNamespace]
        """

        posts = await self.posts(
            session=session,
            status=status,
            limit=posts_limit,
            sort=sort,
            timeframe=timeframe,
        )
        all_comments: List = []
        found_comments: List = []
        for post in posts:
            post = Post(
                id=post.get("id"),
                subreddit=self._name,
                time_format=self._time_format,
            )

            comments = await post.comments(
                session=session, limit=comments_per_post, status=status
            )

            all_comments.extend(comments)

        pattern = rf"(?i)\b{re.escape(query)}\b"
        regex: re.Pattern = re.compile(pattern, re.IGNORECASE)

        for comment in all_comments:
            match: re.Match = regex.search(comment.body)
            if match:
                found_comments.append(comment)

        if found_comments:
            return found_comments

    async def search_posts(
        self,
        session: ClientSession,
        query: str,
        limit: int,
        sort: SORT_CRITERION = "all",
        timeframe: TIMEFRAME = "all",
        status: Optional[Status] = None,
    ) -> List[SimpleNamespace]:
        """
        Asynchronously get posts that contain the specified query string from a subreddit.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param query: Search query.
        :type query: str
        :param limit: Maximum number of posts to return.
        :type limit: int
        :param sort: Sort criterion for the posts.
        :type sort: str
        :param timeframe: Timeframe from which to get posts.
        :type timeframe: Literal[str]
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing post data.
        :rtype: List[SimpleNamespace]
        """

        search_results = await api.get_posts_or_comments(
            session=session,
            kind="search_from_a_subreddit",
            status=status,
            subreddit=self._name,
            query=query,
            limit=limit,
            sort=sort,
            timeframe=timeframe,
        )

        parsed_posts = self._parse.posts(search_results)

        return parsed_posts if search_results else [SimpleNamespace]

    async def wiki_pages(
        self, session: ClientSession, status: Optional[Status] = None
    ) -> List[str]:
        """
        Asynchronously get a subreddit's wiki pages.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of strings, each representing a wiki page.
        :rtype: List[str]
        """

        if status:
            status.update(
                f"Retrieving wiki pages from subreddit ({self._name})",
            )

        pages = await api.send_request(
            endpoint=f"{api.subreddit_endpoint}/{self._name}/wiki/pages.json",
            session=session,
        )

        return pages.get("data")

    async def wiki_page(
        self,
        page_name: str,
        session: ClientSession,
        status: Optional[Status] = None,
    ) -> SimpleNamespace:
        """
        Asynchronously get a subreddit's specified wiki page data.

        :param page_name: Wiki page to get data from.
        :type page_name: str
        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A `SimpleNamespace` object containing the parsed wiki page data.
        :rtype: SimpleNamespace
        """

        wiki_page = await api.get_entity(
            session=session,
            kind="wikipage",
            status=status,
            page_name=page_name,
            subreddit=self._name,
        )

        parsed_wiki_page = self._parse.wiki_page(wiki_page)

        return parsed_wiki_page if wiki_page else SimpleNamespace


class Subreddits:
    """Represents Reddit subreddits and provides methods for getting related data."""

    def __init__(self, time_format: TIME_FORMAT = "locale"):
        """
        Initialises the `Subreddits()` instance for getting `all`, `default`, `new` and `popular` subreddits.

        :param time_format: Time format of the output data. Use `concise` for a human-readable
                        time difference, or `locale` for a localized datetime string. Defaults to `locale`.
        :type time_format: Literal["concise", "locale"]
        """

        self._parse = Parse(time_format=time_format)

    async def all(
        self,
        session: ClientSession,
        limit: int,
        timeframe: TIMEFRAME = "all",
        status: Optional[Status] = None,
    ) -> List[SimpleNamespace]:
        """
        Asynchronously get all subreddits.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param limit: Maximum number of subreddits to return.
        :type limit: int
        :param timeframe: Timeframe from which to get all subreddits.
        :type timeframe: Literal[str]
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing subreddit data.
        :rtype: List[SimpleNamespace]

        Note:
            Items will most likely be limited to 1000, per Reddit's API policy.
        """

        all_subreddits = await api.get_subreddits(
            session=session,
            kind="all",
            status=status,
            limit=limit,
            timeframe=timeframe,
        )

        parsed_subreddits = self._parse.subreddits(all_subreddits)

        return parsed_subreddits if all_subreddits else [SimpleNamespace]

    async def default(
        self,
        limit: int,
        session: ClientSession,
        status: Optional[Status] = None,
    ) -> List[SimpleNamespace]:
        """
        Asynchronously get default subreddits.

        :param limit: Maximum number of subreddits to return.
        :type limit: int
        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing subreddit data.
        :rtype: List[SimpleNamespace]
        """

        default_subreddits = await api.get_subreddits(
            session=session,
            kind="default",
            status=status,
            timeframe="all",
            limit=limit,
        )

        parsed_subreddits = self._parse.subreddits(default_subreddits)

        return parsed_subreddits if default_subreddits else [SimpleNamespace]

    async def new(
        self,
        session: ClientSession,
        limit: int,
        timeframe: TIMEFRAME = "all",
        status: Optional[Status] = None,
    ) -> List[SimpleNamespace]:
        """
        Asynchronously get new subreddits.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param limit: Maximum number of subreddits to return.
        :type limit: int
        :param timeframe: Timeframe from which to get new subreddits.
        :type timeframe: Literal[str]
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing subreddit data.
        :rtype: List[SimpleNamespace]
        """
        new_subreddits = await api.get_subreddits(
            session=session,
            status=status,
            kind="new",
            limit=limit,
            timeframe=timeframe,
        )

        parsed_subreddits = self._parse.subreddits(new_subreddits)

        return parsed_subreddits if new_subreddits else [SimpleNamespace]

    async def popular(
        self,
        session: ClientSession,
        limit: int,
        timeframe: TIMEFRAME = "all",
        status: Optional[Status] = None,
    ) -> List[SimpleNamespace]:
        """
        Asynchronously get popular subreddits.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param limit: Maximum number of subreddits to return.
        :type limit: int
        :param timeframe: Timeframe from which to get popular subreddits.
        :type timeframe: Literal[str]
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing subreddit data.
        :rtype: List[SimpleNamespace]
        """

        popular_subreddits = await api.get_subreddits(
            session=session,
            kind="popular",
            status=status,
            limit=limit,
            timeframe=timeframe,
        )

        parsed_subreddits = self._parse.subreddits(popular_subreddits)

        return parsed_subreddits if popular_subreddits else [SimpleNamespace]


class User:
    """Represents a Reddit user and provides methods for getting data from the specified user."""

    def __init__(self, name: str, time_format: TIME_FORMAT = "locale"):
        """
        Initialises a `User()` instance for getting a user's `profile`, `posts` and `comments` data.

        :param name: Username to get data from.
        :type name: str
        :param time_format: Time format of the output data. Use "concise" for a human-readable
                        time difference, or "locale" for a localized datetime string. Defaults to "locale".
        :type time_format: Literal["concise", "locale"]
        """

        self._name = name
        self._parse = Parse(time_format=time_format)

    async def comments(
        self,
        session: ClientSession,
        limit: int,
        sort: SORT_CRITERION = "all",
        timeframe: TIMEFRAME = "all",
        status: Optional[Status] = None,
    ) -> List[SimpleNamespace]:
        """
        Asynchronously get a user's comments.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param limit: Maximum number of comments to return.
        :type limit: int
        :param sort: Sort criterion for the comments.
        :type sort: str
        :param timeframe: Timeframe from which tyo get comments.
        :type timeframe: Literal[str]
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing comment data.
        :rtype: List[SimpleNamespace]
        """

        user_comments = await api.get_posts_or_comments(
            session=session,
            kind="comments_from_a_user",
            status=status,
            username=self._name,
            limit=limit,
            sort=sort,
            timeframe=timeframe,
        )

        parsed_comments = self._parse.comments(user_comments)

        return parsed_comments if user_comments else [SimpleNamespace]

    async def moderated_subreddits(
        self, session: ClientSession, status: Optional[Status] = None
    ) -> List[SimpleNamespace]:
        """
        Asynchronously get subreddits moderated by user.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing subreddit data.
        :rtype: List[SimpleNamespace]
        """

        subreddits = await api.get_subreddits(
            session=session,
            kind="user_moderated",
            status=status,
            username=self._name,
            limit=0,
        )

        parsed_subreddits = self._parse.subreddits(subreddits)

        return parsed_subreddits if subreddits else [SimpleNamespace]

    async def overview(
        self,
        limit: int,
        session: ClientSession,
        status: Optional[Status] = None,
    ) -> List[SimpleNamespace]:
        """
        Asynchronously get a user's most recent comments.

        :param limit: Maximum number of comments to return.
        :type limit: int
        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing data about a recent comment.
        :rtype: List[SimpleNamespace]
        """

        user_overview = await api.get_posts_or_comments(
            session=session,
            kind="overview_of_a_user",
            status=status,
            username=self._name,
            limit=limit,
        )

        parsed_overview = self._parse.comments(user_overview)

        return parsed_overview if user_overview else [SimpleNamespace]

    async def posts(
        self,
        session: ClientSession,
        limit: int,
        sort: SORT_CRITERION = "all",
        timeframe: TIMEFRAME = "all",
        status: Optional[Status] = None,
    ) -> List[SimpleNamespace]:
        """
        Asynchronously get a user's posts.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param limit: Maximum number of posts to return.
        :type limit: int
        :param sort: Sort criterion for the posts.
        :type sort: str
        :param timeframe: Timeframe from which to get posts.
        :type timeframe: Literal[str]
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing post data.
        :rtype: List[SimpleNamespace]
        """

        user_posts = await api.get_posts_or_comments(
            session=session,
            kind="posts_from_a_user",
            status=status,
            username=self._name,
            limit=limit,
            sort=sort,
            timeframe=timeframe,
        )

        parsed_posts = self._parse.posts(user_posts)

        return parsed_posts if user_posts else [SimpleNamespace]

    async def profile(
        self, session: ClientSession, status: Optional[Status] = None
    ) -> SimpleNamespace:
        """
        Asynchronously get a user's profile data.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A SimpleNamespace object containing user profile data.
        :rtype: SimpleNamespace
        """

        user_profile = await api.get_entity(
            username=self._name, kind="user", status=status, session=session
        )

        parsed_profile = self._parse.user(user_profile)

        return parsed_profile if user_profile else SimpleNamespace

    async def search_posts(
        self,
        query: str,
        limit: int,
        session: ClientSession,
        sort: SORT_CRITERION = "all",
        timeframe: TIMEFRAME = "all",
        status: Optional[Status] = None,
    ) -> List[SimpleNamespace]:
        """
        Asynchronously get a user's posts that match with the specified search query.

        :param query: Search query.
        :type query: str
        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param limit: Maximum number of posts to search from.
        :type limit: int
        :param sort: Sort criterion for the posts.
        :type sort: str
        :param timeframe: Timeframe from which to get posts.
        :type timeframe: Literal[str]
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing post data.
        :rtype: List[SimpleNamespace]
        """

        pattern = rf"(?i)\b{re.escape(query)}\b"
        regex: re.Pattern = re.compile(pattern, re.IGNORECASE)

        user_posts = await api.get_posts_or_comments(
            session=session,
            kind="posts_from_a_user",
            status=status,
            username=self._name,
            limit=limit,
            sort=sort,
            timeframe=timeframe,
        )
        found_posts = []

        for post in user_posts:
            post_data = post.get("data")

            match: re.Match = regex.search(post_data.get("title", "")) or regex.search(
                post_data.get("selftext", "")
            )

            if match:
                found_posts.append(post)

        parsed_post = self._parse.posts(found_posts)

        return parsed_post if found_posts else [SimpleNamespace]

    async def search_comments(
        self,
        query: str,
        limit: int,
        session: ClientSession,
        sort: SORT_CRITERION = "all",
        timeframe: TIMEFRAME = "all",
        status: Optional[Status] = None,
    ) -> List[SimpleNamespace]:
        """
        Asynchronously get a user's comments that contain the specified search query.

        :param query: Search query.
        :type query: str
        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :type session: aiohttp.ClientSession
        :param limit: Maximum number of comments to search from.
        :type limit: int
        :param sort: Sort criterion for the comments.
        :type sort: str
        :param timeframe: Timeframe from which to get comments.
        :type timeframe: Literal[str]
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing comment data.
        :rtype: List[SimpleNamespace]
        """

        pattern: str = rf"(?i)\b{re.escape(query)}\b"
        regex: re.Pattern = re.compile(pattern, re.IGNORECASE)

        user_comments = await api.get_posts_or_comments(
            session=session,
            kind="comments_from_a_user",
            status=status,
            username=self._name,
            limit=limit,
            sort=sort,
            timeframe=timeframe,
        )
        found_comments = []

        for comment in user_comments:
            match = regex.search(comment.get("data", {}).get("body", ""))
            if match:
                found_comments.append(comment)

        parsed_comments = self._parse.comments(found_comments)

        return parsed_comments if user_comments else [SimpleNamespace]

    async def top_subreddits(
        self,
        session: ClientSession,
        top_n: int,
        limit: int,
        filename: str = None,
        timeframe: TIMEFRAME = "all",
        status: Optional[Status] = None,
    ) -> Union[List[tuple[str, int]], None]:
        """
        Asynchronously get a user's top n subreddits based on subreddit frequency in n posts and saves the analysis to a file.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param top_n: Communities arranging number.
        :type top_n: int
        :param limit: Maximum number of posts to scrape.
        :type limit: int
        :param filename: Filename to which the analysis will be saved.
        :type filename: str
        :param timeframe: Timeframe from which to get posts.
        :type timeframe: Literal[str]
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        """

        posts = await api.get_posts_or_comments(
            session=session,
            kind="posts_from_a_user",
            status=status,
            username=self._name,
            limit=limit,
            timeframe=timeframe,
        )

        if posts:
            # Extract subreddit names
            subreddits = [post.get("data", {}).get("subreddit") for post in posts]

            # Count the occurrences of each subreddit
            subreddit_counts: Counter = Counter(subreddits)

            # Get the most common subreddits
            top_subreddits: List[tuple[str, int]] = subreddit_counts.most_common(top_n)

            # Prepare data for plotting
            subreddit_names = [subreddit[0] for subreddit in top_subreddits]
            subreddit_frequencies = [subreddit[1] for subreddit in top_subreddits]

            if visualisation_dependency_installed:
                plot_bar_chart(
                    data=dict(zip(subreddit_names, subreddit_frequencies)),
                    title=f"top {top_n}/{limit} subreddits analysis",
                    xlabel="Subreddits",
                    ylabel="Frequency",
                    figure_size=(top_n + 10, 5),
                    colours=["lightblue"] * top_n,
                    filename=f"{filename + '_' if filename else ''}top_{top_n}_of_{limit}_subreddits",
                )
            else:
                return top_subreddits

    async def username_available(
        self, session: ClientSession, status: Optional[Status] = None
    ) -> bool:
        """
        Checks if the given username is available or taken.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: `True` if the given username is available. Otherwise, `False`.
        :rtype: bool
        """

        if status:
            status.update(f"Checking username availability: {self._name}")

        response: bool = await api.send_request(
            session=session,
            endpoint=api.endpoint("username_available"),
            params={"user": self._name},
        )

        return response


class Users:
    """Represents Reddit users and provides methods for getting related data."""

    def __init__(self, time_format: TIME_FORMAT = "locale"):
        """
        Initialises the `Users()` instance for getting `new`, `popular` and `all` users.

        :param time_format: Time format of the output data. Use `concise` for a human-readable
                        time difference, or `locale` for a localized datetime string. Defaults to `locale`.
        :type time_format: Literal["concise", "locale"]
        """  #
        self._parse = Parse(time_format=time_format)

    async def new(
        self,
        session: ClientSession,
        limit: int,
        timeframe: TIMEFRAME = "all",
        status: Optional[Status] = None,
    ) -> List[SimpleNamespace]:
        """
        Asynchronously get new users.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param limit: Maximum number of new users to return.
        :type limit: int
        :param timeframe: Timeframe from which to get new posts.
        :type timeframe: Literal[str]
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing a user's data.
        :rtype: List[SimpleNamespace]
        """

        new_users = await api.get_users(
            session=session,
            status=status,
            kind="new",
            limit=limit,
            timeframe=timeframe,
        )

        parsed_users = self._parse.users(new_users)

        return parsed_users if new_users else [SimpleNamespace]

    async def popular(
        self,
        session: ClientSession,
        limit: int,
        timeframe: TIMEFRAME = "all",
        status: Optional[Status] = None,
    ) -> List[SimpleNamespace]:
        """
        Asynchronously get popular users.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param limit: Maximum number of popular users to return.
        :type limit: int
        :param timeframe: Timeframe from which to get popular posts.
        :type timeframe: Literal[str]
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing a user's data.
        :rtype: List[SimpleNamespace]
        """

        popular_users = await api.get_users(
            session=session,
            kind="popular",
            status=status,
            limit=limit,
            timeframe=timeframe,
        )

        parsed_users = self._parse.users(popular_users)

        return parsed_users if popular_users else [SimpleNamespace]

    async def all(
        self,
        session: ClientSession,
        limit: int,
        timeframe: TIMEFRAME = "all",
        status: Optional[Status] = None,
    ) -> List[SimpleNamespace]:
        """
        Asynchronously get all users.

        :param limit: Maximum number of all users to return.
        :type limit: int
        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param timeframe: Timeframe from which to get all posts.
        :type timeframe: Literal[str]
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing a user's data.
        :rtype: List[SimpleNamespace]
        """

        all_users = await api.get_users(
            session=session,
            kind="all",
            status=status,
            limit=limit,
            timeframe=timeframe,
        )

        parsed_users = self._parse.users(all_users)

        return parsed_users if all_users else [SimpleNamespace]


# -------------------------------- END ----------------------------------------- #



import os
from types import SimpleNamespace
from typing import Union, Literal, List, Tuple, Dict

import pandas as pd

from .shared import notify

__all__ = [
    "create_dataframe",
    "export_dataframe",
    "plot_bar_chart",
    "EXPORT_FORMATS",
    "visualisation_dependency_installed",
]

EXPORT_FORMATS = Literal["csv", "html", "json", "xml"]

try:
    import matplotlib.pyplot as plt

    visualisation_dependency_installed: bool = True
except ImportError:
    visualisation_dependency_installed = False


def plot_bar_chart(
    data: Dict[str, int],
    title: str,
    xlabel: str,
    ylabel: str,
    colours: List[str],
    filename: str,
    figure_size: Tuple[int, int] = (10, 5),
):
    """
    Plots a bar chart for the given data.

    :param data: A dictionary where keys are the categories and values are the counts or frequencies.
    :type data: list[str, int]
    :param title: The title of the plot.
    :type title: str
    :param xlabel: The label for the x-axis.
    :type xlabel: str
    :param ylabel: The label for the y-axis.
    :type ylabel: str
    :param colours: A list of colours to use for the bars in the chart.
    :type colours: list[str]
    :param filename: The name of the file where the plot will be saved.
    :type filename: str
    :param figure_size: The size of the figure (width, height). Defaults to (10, 5).
    :type figure_size: tuple[int, int]
    """

    if visualisation_dependency_installed:
        plt.figure(figsize=figure_size)
        plt.bar(list(data.keys()), list(data.values()), color=colours)
        plt.title(title)
        plt.xlabel(xlabel)
        plt.ylabel(ylabel)
        plt.savefig(f"{filename}.png")

        notify.ok(f"{title} saved to [link file://{filename}.png]{filename}.png")


def create_dataframe(
    data: Union[SimpleNamespace, List[SimpleNamespace], List[Tuple[str, int]]],
) -> pd.DataFrame:
    """
    Makes a Pandas dataframe from the provided data.

    :param data: Data to be converted.
    :type data: Union[SimpleNamespace, List[SimpleNamespace], List[Tuple[str, int]]]
    :return: A pandas DataFrame constructed from the provided data. Excludes any 'raw_data'
             column from the dataframe.
    :rtype: pd.DataFrame
    """

    if isinstance(data, SimpleNamespace):
        # Transform each attribute of the object into a dictionary entry
        transformed_data = [
            {"key": key, "value": value} for key, value in data.__dict__.items()
        ]

    # Convert a list of SimpleNamespace objects to a list of dictionaries
    elif isinstance(data, List) and all(
        isinstance(item, SimpleNamespace) for item in data
    ):
        # Each object in the list is transformed to its dictionary representation
        transformed_data = [item.__dict__ for item in data]
    else:
        transformed_data = data

    # Set pandas display option to show all rows
    pd.set_option("display.max_rows", None)

    # Create a DataFrame from the transformed data
    dataframe = pd.DataFrame(transformed_data)

    return dataframe.dropna(axis=1, how="all")


def export_dataframe(
    dataframe: pd.DataFrame,
    filename: str,
    directory: str,
    formats: List[EXPORT_FORMATS],
):
    """
    Exports a Pandas dataframe to specified file formats.

    :param dataframe: Pandas dataframe to export.
    :type dataframe: pandas.DataFrame
    :param filename: Name of the file to which the dataframe will be exported.
    :type filename: str
    :param directory: Directory to which the dataframe files will be saved.
    :type directory: str
    :param formats: A list of file formats to which the data will be exported.
    :type formats: List[Literal]
    """

    file_mapping: Dict = {
        "csv": lambda: dataframe.to_csv(
            os.path.join(directory, "csv", f"{filename}.csv"), encoding="utf-8"
        ),
        "html": lambda: dataframe.to_html(
            os.path.join(directory, "html", f"{filename}.html"),
            escape=False,
            encoding="utf-8",
        ),
        "json": lambda: dataframe.to_json(
            os.path.join(directory, "json", f"{filename}.json"),
            force_ascii=False,
            indent=4,
        ),
        "xml": lambda: dataframe.to_xml(
            os.path.join(directory, "xml", f"{filename}.xml"),
            parser="etree",
            encoding="utf-8",
        ),
    }

    for file_format in formats:
        if file_format in file_mapping:
            filepath: str = os.path.join(
                directory, file_format, f"{filename}.{file_format}"
            )
            file_mapping.get(file_format)()
            notify.ok(
                f"{get_file_size(file_path=filepath)} written to [link file://{filepath}]{filepath}"
            )
        else:
            continue


def get_file_size(file_path: str) -> str:
    """
    Gets a file size and puts it in human-readable form.

    :param file_path: Path to target file.
    :type file_path: str
    :return: A human-readable form of the file size.
    :rtype: str
    """

    file_size_bytes: int = os.path.getsize(file_path)
    units: list = ["B", "KB", "MB", "GB", "TB", "PB"]

    unit_index: int = 0

    while file_size_bytes >= 1024 and unit_index < len(units) - 1:
        file_size_bytes /= 1024
        unit_index += 1

    return f"{file_size_bytes:.2f} {units[unit_index]}"


# -------------------------------- END ----------------------------------------- #



import asyncio
import time
from random import randint
from typing import Callable, Literal, Union, Optional, List, Dict

from aiohttp import ClientSession
from karmakaze import Sanitise
from rich.status import Status

__all__ = ["Api", "SORT_CRITERION", "TIMEFRAME", "TIME_FORMAT"]

SORT_CRITERION = Literal["controversial", "new", "top", "best", "hot", "rising", "all"]
TIMEFRAME = Literal["hour", "day", "week", "month", "year", "all"]
TIME_FORMAT = Literal["concise", "locale"]


class Api:
    """Represents the Knew Karma API and provides methods for getting various data from the Reddit API."""

    def __init__(self, headers: Optional[Dict] = None):
        self._headers = headers
        self._sanitise = Sanitise()

    @staticmethod
    def endpoint(
        kind: Literal[
            "base",
            "user",
            "users",
            "subreddit",
            "subreddits",
            "reddit_status",
            "reddit_status_components",
            "username_available",
        ]
    ) -> str:
        """
        A static method that contains endpoints for the specified `kind` of data.

        :param kind: Kind of data to get endpoint from.
        :type kind: Literal[str]
        :return: An endpoint of the specified `kind`.
        :rtype: str
        """
        base = "https://www.reddit.com"
        endpoint_map = {
            "base": base,
            "user": f"{base}/user",
            "users": f"{base}/users",
            "subreddit": f"{base}/r",
            "subreddits": f"{base}/subreddits",
            "reddit_status": "https://www.redditstatus.com/api/v2/status.json",
            "reddit_status_components": "https://www.redditstatus.com/api/v2/components.json",
            "username_available": f"{base}/api/username_available.json",
        }

        return endpoint_map.get(kind)

    async def send_request(
        self, session: ClientSession, endpoint: str, params: Optional[Dict] = None
    ) -> Union[Dict, List, bool, None]:
        """
        Asynchronously sends a GET request to the specified API endpoint and returns JSON or list response.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param endpoint: The API endpoint to fetch data from.
        :type endpoint: str
        :param params: A dictionary containing requests parameters. Defaults to None.
        :type params: Dict
        :return: JSON data as a dictionary or list or a boolean value. Raises an exception if fetching fails.
        :rtype: Union[Dict, List, bool]
        :raise Exception: If any is encountered.
        """

        try:
            async with session.get(
                url=endpoint, headers=self._headers, params=params
            ) as response:
                response.raise_for_status()
                response_data: Union[Dict, List] = await response.json()
                return response_data

        except Exception as error:
            raise error

    async def _paginate_items(
        self,
        session: ClientSession,
        sanitiser: Callable,
        limit: int,
        status: Optional[Status] = None,
        **kwargs: Union[str, bool, Dict],
    ) -> List[Dict]:
        """
        Asynchronously fetches and processes data in a paginated manner
        from a specified endpoint until the specified limit
        of items is reached or there are no more items to fetch. It uses a specified processing function
        to handle the data from each request, ensuring no duplicates are returned.

        :param session: An Aiohttp session to use for the request.
        :type session: aiohttp.ClientSession
        :param sanitiser: A callable used to sanitise response data.
        :type sanitiser: Callable
        :param limit: Maximum number of results to return.
        :type limit: int
        :return: A list of dict objects, each containing paginated data.
        :rtype: List[Dict]
        """

        # Initialise an empty list to store all items across paginated requests.
        all_items: List = []
        # Initialise the ID of the last item fetched to None (used for pagination).
        last_item_id = None

        params: Dict = kwargs.get("params")

        # Continue fetching data until the limit is reached or no more items are available.
        while len(all_items) < limit:
            # Make an asynchronous request to the endpoint.
            response = await self.send_request(
                session=session,
                endpoint=kwargs.get("endpoint"),
                params=(
                    params.update({"after": last_item_id, "count": len(all_items)})
                    if last_item_id
                    else params
                ),
            )

            # If the request is for post comments, handle the response accordingly.
            if kwargs.get("is_comments_from_a_post"):
                items = []  # Initialise a list to store fetched items.
                more_items_ids = []  # Initialise a list to store IDs from "more" items.

                # Iterate over the children in the response to extract comments or "more" items.
                for item in response[1].get("data").get("children"):
                    if self._sanitise.kind(item) == "t1":
                        sanitised_item = sanitiser(item)
                        # If the item is a comment (kind == "t1"), add it to the items list.
                        items.append(sanitised_item)
                    elif self._sanitise.kind(item) == "more":
                        # If the item is of kind "more", extract the IDs for additional comments.
                        more_items_ids.extend(item)

                # If there are more items to fetch (kind == "more"), make additional requests.
                if more_items_ids:
                    await self._paginate_more_items(
                        session=session,
                        fetched_items=items,
                        more_items_ids=more_items_ids,
                        endpoint=kwargs.get("endpoint"),
                    )

            # If not handling comments, simply extract the items from the response.
            items = sanitiser(response)

            # If no items are found, break the loop as there's nothing more to fetch.
            if not items:
                break

            # Determine how many more items are needed to reach the limit.
            items_to_limit = limit - len(all_items)

            # Add the processed items to the all_items list, up to the specified limit.
            all_items.extend(items[:items_to_limit])

            # Update the last_item_id to the ID of the last fetched item for pagination.
            last_item_id = (
                self._sanitise.pagination_id(response=response[1])
                if kwargs.get("is_post_comments")
                else self._sanitise.pagination_id(response=response)
            )

            # If we've reached the specified limit, break the loop.
            if len(all_items) == limit:
                break

            # Introduce a random sleep duration between 1 and 5 seconds to avoid rate-limiting.
            sleep_duration = randint(1, 5)

            # If a status object is provided, use it to display a countdown timer.
            if status:
                await self._pagination_countdown_timer(
                    status=status,
                    duration=sleep_duration,
                    current_count=len(all_items),
                    overall_count=limit,
                )
            else:
                # Otherwise, just sleep for the calculated duration.
                await asyncio.sleep(sleep_duration)

        # Return the list of all fetched and processed items.
        return all_items

    async def _paginate_more_items(
        self,
        session: ClientSession,
        more_items_ids: List[str],
        endpoint: str,
        fetched_items: List[Dict],
    ):
        for more_id in more_items_ids:
            # Construct the endpoint for each additional comment ID.
            more_endpoint = f"{endpoint}&comment={more_id}"
            # Make an asynchronous request to fetch the additional comments.
            more_response = await self.send_request(
                session=session, endpoint=more_endpoint
            )
            # Extract the items (comments) from the response.
            more_items, _ = self._sanitise.comments(
                response=more_response[1].get("data", {}).get("children", [])
            )

            # Add the fetched items to the main items list.
            fetched_items.extend(more_items)

    @staticmethod
    async def _pagination_countdown_timer(
        status: Status, duration: int, current_count: int, overall_count: int
    ):
        """
        A static method handles the live countdown during pagination, updating the status bar with the remaining time.

        :param status: A Status object used to display the countdown.
        :type status: rich.status.Status
        :param duration: The duration for which to run the countdown.
        :type duration: int
        :param current_count: Current number of items fetched.
        :type current_count: int
        :param overall_count: Overall number of items to fetch.
        :type overall_count: int
        """

        from .tools.shared import style

        end_time: float = time.time() + duration
        while time.time() < end_time:
            remaining_time: float = end_time - time.time()
            remaining_seconds: int = int(remaining_time)
            remaining_milliseconds: int = int(
                (remaining_time - remaining_seconds) * 100
            )

            status.update(
                f"{style.cyan}{current_count}{style.reset} (of {style.cyan}{overall_count}{style.reset}) items retrieved so far. "
                f"Resuming in {style.cyan}{remaining_seconds}.{remaining_milliseconds:02}{style.reset} seconds",
            )
            await asyncio.sleep(0.01)  # Sleep for 10 milliseconds

    async def check_reddit_status(
        self, session: ClientSession, status: Optional[Status] = None
    ):
        """
        Asynchronously checks Reddit API and infrastructure status.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param status: An optional `Status` object for displaying status messages.
        :type status: Optional[rich.status.Status]
        """

        from rich.table import Table
        from rich import box
        from .tools.shared import console, notify, style

        if status:
            status.update(
                f"Checking Reddit {style.bold}API/Infrastructure{style.reset} status"
            )

        status_response: Dict = await self.send_request(
            endpoint=self.endpoint(kind="reddit_status"), session=session
        )

        indicator = status_response.get("status").get("indicator")
        description = status_response.get("status").get("description")
        if description:
            if indicator == "none":
                notify.ok(description)
            else:
                notify.warning(
                    f"{description} ({style.yellow}{indicator}{style.reset})"
                )
                if status:
                    status.update("Getting status components")
                status_components: Dict = await self.send_request(
                    endpoint=self.endpoint(kind="reddit_status_components"),
                    session=session,
                )

                if isinstance(status_components, Dict):
                    components: List[Dict] = status_components.get("components")
                    if components:
                        table = Table(
                            title="Reddit API/Infrastructure Status",
                            box=box.ROUNDED,
                        )
                        table.add_column(
                            "Component", justify="right", no_wrap=True, style="dim"
                        )
                        table.add_column("Description")
                        table.add_column("Status", justify="right")
                        table.add_column("Updated at", style=style.cyan.strip("[,]"))

                        for component in components:
                            component_name = component.get("name")
                            component_status = component.get("status")
                            component_description = component.get("description")
                            component_update_date = component.get("updated_at")

                            component_colour = (
                                style.green
                                if component_status == "operational"
                                else style.red
                            )
                            table.add_row(
                                component_name,
                                component_description,
                                f"{component_colour}{component_status}{style.reset}",
                                component_update_date,
                            )

                        console.print(table)

    async def get_entity(
        self,
        session: ClientSession,
        kind: Literal["comment", "post", "subreddit", "user", "wikipage"],
        status: Optional[Status] = None,
        **kwargs: str,
    ) -> Dict:
        """
        Asynchronously gets data from the specified entity.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param kind: The type of entity to get data from
        :type kind: str
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: rich.status.Status
        :return: A dictionary containing a specified entity's data.
        :rtype: Dict
        """

        username = kwargs.get("username")
        subreddit = kwargs.get("subreddit")
        post_id = kwargs.get("id")
        comment_id = kwargs.get("comment_id")

        entity_mapping = {
            """
            "comment": {
                "endpoint": f"{self.endpoint(kind='subreddit')}",
                "sanitiser": lambda data: self._sanitise.comment(data),
            },
            """
            "post": {
                "endpoint": f"{self.endpoint(kind='subreddit')}/{subreddit}"
                f"/comments/{post_id}.json",
                "sanitiser": lambda data: self._sanitise.post(data),
            },
            "user": {
                "endpoint": f"{self.endpoint(kind='user')}/{username}/about.json",
                "sanitiser": lambda data: self._sanitise.subreddit_or_user(data),
            },
            "subreddit": {
                "endpoint": f"{self.endpoint(kind='subreddit')}/{subreddit}/about.json",
                "sanitiser": lambda data: self._sanitise.subreddit_or_user(data),
            },
            "wikipage": {
                "endpoint": f"{self.endpoint(kind='subreddit')}/{subreddit}/wiki/{kwargs.get('page_name')}.json",
                "sanitiser": lambda data: self._sanitise.wiki_page(data),
            },
        }

        if status:
            status.update(
                f"Retrieving {kind} data",
            )

        endpoint = entity_mapping.get(kind).get("endpoint")
        sanitiser = entity_mapping.get(kind).get("sanitiser")

        response = await self.send_request(endpoint=endpoint, session=session)
        sanitised_response = sanitiser(response)

        return sanitised_response

    async def get_posts_or_comments(
        self,
        session: ClientSession,
        kind: Literal[
            "best",
            "controversial",
            "front_page",
            "new",
            "popular",
            "rising",
            "posts_from_a_subreddit",
            "search_from_a_subreddit",
            "posts_from_a_user",
            "overview_of_a_user",
            "comments_from_a_user",
            "comments_from_a_post",
        ],
        limit: int,
        timeframe: TIMEFRAME = "all",
        sort: SORT_CRITERION = "all",
        status: Optional[Status] = None,
        **kwargs: str,
    ) -> List[Dict]:
        """
        Asynchronously gets a specified number of posts or comments, with a specified sorting criterion, from the specified source.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param kind: The kind of posts to be fetched.
        :type kind: str
        :param limit: Maximum number of posts to get.
        :type limit: int
        :param sort: Posts' sort criterion.
        :type sort: str
        :param timeframe: Timeframe from which to get posts.
        :type timeframe: Literal
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: rich.status.Status
        :return: A list of dictionaries, each containing post data.
        :rtype: List[Dict]
        """

        username = kwargs.get("username")
        subreddit = kwargs.get("subreddit")

        posts_or_comments_mapping = {
            "best": f"{self.endpoint(kind='base')}/r/{kind}.json",
            "controversial": f"{self.endpoint(kind='base')}/r/{kind}.json",
            "front_page": f"{self.endpoint(kind='base')}/.json",
            "new": f"{self.endpoint(kind='base')}/new.json",
            "popular": f"{self.endpoint(kind='base')}/r/{kind}.json",
            "rising": f"{self.endpoint(kind='base')}/r/{kind}.json",
            "posts_from_a_subreddit": f"{self.endpoint(kind='subreddit')}/{subreddit}.json",
            "posts_from_a_user": f"{self.endpoint(kind='user')}/{username}/submitted.json",
            "overview_of_a_user": f"{self.endpoint(kind='user')}/{username}/overview.json",
            "comments_from_a_user": f"{self.endpoint(kind='user')}/{username}/comments.json",
            "comments_from_a_post": f"{self.endpoint(kind='subreddit')}/{subreddit}"
            f"/comments/{kwargs.get('id')}.json",
            "search_from_a_subreddit": f"{self.endpoint(kind='subreddit')}/{subreddit}"
            f"/search.json",
        }

        if status:
            status.update(f"Retrieving {limit} {kind} (posts/comments)")

        endpoint = posts_or_comments_mapping.get(kind)
        params = {"limit": limit, "sort": sort, "t": timeframe, "raw_json": 1}
        params.update(
            {"q": kwargs.get("query"), "restrict_sr": 1}
            if kind == "search_from_a_subreddit"
            else {}
        )

        sanitiser = (
            self._sanitise.comments
            if kind in ["user_comments", "user_overview", "post_comments"]
            else self._sanitise.posts
        )

        posts = await self._paginate_items(
            session=session,
            endpoint=endpoint,
            params=params,
            status=status,
            sanitiser=sanitiser,
            limit=limit,
            is_comments_from_a_post=True if kind == "comments_from_a_post" else False,
        )

        return posts

    async def get_subreddits(
        self,
        session: ClientSession,
        kind: Literal["all", "default", "new", "popular", "user_moderated"],
        limit: int,
        timeframe: TIMEFRAME = "all",
        status: Optional[Status] = None,
        **kwargs: str,
    ) -> Union[List[Dict], Dict]:
        """
        Asynchronously gets the specified type of subreddits.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param kind: The kind of subreddits to get.
        :type kind: str
        :param limit: Maximum number of subreddits to return.
        :type limit: int
        :param timeframe: Timeframe from which to get subreddits.
        :type timeframe: Literal
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: rich.status.Status
        :return: A list of dictionaries, each containing subreddit data,
            or a single dictionary containing subreddit data.
        :rtype: Union[List[Dict], Dict]
        """

        subreddits_mapping = {
            "all": f"{self.endpoint(kind='subreddits')}.json",
            "default": f"{self.endpoint(kind='subreddits')}/default.json",
            "new": f"{self.endpoint(kind='subreddits')}/new.json",
            "popular": f"{self.endpoint(kind='subreddits')}/popular.json",
            "user_moderated": f"{self.endpoint(kind='user')}/{kwargs.get('username')}/moderated_subreddits.json",
        }

        if status:
            status.update(f"Retrieving {limit} {kind} subreddits")

        endpoint = subreddits_mapping.get(kind, "")
        params = {"raw_json": 1}

        if kind == "user_moderated":
            subreddits = await self.send_request(
                endpoint=endpoint,
                session=session,
            )
        else:
            params.update({"limit": limit, "t": timeframe})
            subreddits = await self._paginate_items(
                session=session,
                params=params,
                endpoint=endpoint,
                sanitiser=self._sanitise.subreddits_or_users,
                limit=limit,
                status=status,
            )

        return subreddits

    async def get_users(
        self,
        session: ClientSession,
        kind: Literal["all", "popular", "new"],
        limit: int,
        timeframe: TIMEFRAME = "all",
        status: Optional[Status] = None,
    ) -> List[Dict]:
        """
        Asynchronously gets the specified type of subreddits.

        :param kind: The kind of users to get.
        :type kind: str
        :param limit: Maximum number of users to return.
        :type limit: int
        :param timeframe: Timeframe from which to get users.
        :type timeframe: Literal
        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of dictionaries, each containing user data.
        :rtype: List[Dict]
        """

        users_mapping = {
            "all": f"{self.endpoint(kind='users')}.json",
            "new": f"{self.endpoint(kind='users')}/new.json",
            "popular": f"{self.endpoint(kind='users')}/popular.json",
        }

        if status:
            status.update(f"Retrieving {limit} {kind} users")

        endpoint = users_mapping.get(kind)
        params = {
            "limit": limit,
            "t": timeframe,
        }

        users = await self._paginate_items(
            session=session,
            params=params,
            endpoint=endpoint,
            sanitiser=self._sanitise.subreddits_or_users,
            limit=limit,
            status=status,
        )

        return users

    async def search_entities(
        self,
        session: ClientSession,
        kind: Literal["users", "subreddits", "posts"],
        query: str,
        limit: int,
        sort: SORT_CRITERION = "all",
        status: Optional[Status] = None,
    ) -> List[Dict]:
        """
        Asynchronously searches specified entities that match the specified query.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param kind: The kind of entity to search for.
        :type kind: Literal[str]
        :param query: Search query.
        :type query: str
        :param limit: Maximum number of results to get.
        :type limit: int
        :param sort: Posts' sort criterion.
        :type sort: str
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: rich.status.Status
        :return: A list of dictionaries, each containing search result data.
        :rtype: List[Dict]
        """

        search_mapping = {
            "posts": self.endpoint(kind="base"),
            "subreddits": self.endpoint(kind="subreddits"),
            "users": self.endpoint(kind="users"),
        }

        endpoint = search_mapping.get(kind)
        endpoint += f"/search.json"
        params = {"q": query, "limit": limit, "sort": sort, "raw_json": 1}

        sanitiser = (
            self._sanitise.posts
            if kind == "posts"
            else self._sanitise.subreddits_or_users
        )

        if status:
            status.update(f"Searching for '{query}' in {limit} {kind}")

        search_results = await self._paginate_items(
            session=session,
            params=params,
            endpoint=endpoint,
            status=status,
            sanitiser=sanitiser,
            limit=limit,
        )

        return search_results


# -------------------------------- END ----------------------------------------- #



import os
import subprocess
from typing import Optional, List

import aiohttp
from rich.markdown import Markdown
from rich.prompt import Confirm
from rich.status import Status

from .miscellaneous import make_panel
from .shared import api, console, notify, style
from ..meta import about, version

__all__ = [
    "check_for_updates",
    "is_pypi_package",
    "is_snap_package",
    "update_package",
]

INVALID_PACKAGE_ERROR: str = (
    "The provided package name is not a valid string: {package}"
)


async def check_for_updates(
    session: aiohttp.ClientSession, status: Optional[Status] = None
):
    """
    Asynchronously checks for updates by comparing the current local version with the remote version.

    Assumes version format: major.minor.patch.prefix

    :param session: An `aiohttp.ClientSession` for making the HTTP request.
    :type session: aiohttp.ClientSession
    :param status: An optional `Status` object for displaying status messages.
    :type status: Optional[rich.status.Status]
    """

    if status:
        status.update("Checking for updates")

    # Make a GET request to GitHub to get the project's latest release.
    response = await api.send_request(
        endpoint=f"https://api.github.com/repos/{about.author[1]}/{about.package}/releases/latest",
        session=session,
    )

    if response.get("tag_name"):
        remote_version_str = response.get("tag_name")
        markup_release_notes = response.get("body")

        # Splitting the version strings into components
        local_version_parts = [
            version.major[0],
            version.minor[0],
            version.patch[0],
        ]
        remote_version_parts: List = remote_version_str.split(".")

        local_major: int = local_version_parts[0]
        local_minor: int = local_version_parts[1]
        local_patch: int = local_version_parts[2]

        remote_major = int(remote_version_parts[0])
        remote_minor = int(remote_version_parts[1])
        remote_patch = int(remote_version_parts[2])

        update_level = ""

        # Check for differences in version parts
        if remote_major != local_major:
            update_level = f"{style.red}{version.major[1]}{style.reset}"

        elif remote_minor != local_minor:
            update_level = f"{style.yellow}{version.minor[1]}{style.reset}"

        elif remote_patch != local_patch:
            update_level = f"{style.green}{version.patch[1]}{style.reset}"

        if update_level:
            markdown_release_notes = Markdown(markup=markup_release_notes)
            make_panel(
                title=f"{style.bold}{update_level} Update Available ({style.cyan}{remote_version_str}{style.reset}){style.reset}",
                content=markdown_release_notes,
                subtitle=f"{style.bold}{style.italic}Thank you, for using {about.name}!{style.reset}{style.reset} ❤️ ",
            )

            # Skip auto-updating of the snap package and containerised variant.
            if is_docker_container() or is_snap_package(package=about.package):
                pass
            else:
                status.stop()
                if Confirm.ask(
                    f"{style.bold}Would you like to get these updates?{style.reset}",
                    case_sensitive=False,
                    default=False,
                    console=console,
                ):
                    update_package(package=about.package, status=status)
                else:
                    status.start()
        else:
            notify.ok(message=f"Up-to-date ({version.full})")


def is_docker_container() -> bool:
    """
    Determines if the program is running inside a Docker container.

    :return: True if running inside a Docker container, otherwise False.
    :rtype: bool
    """
    return os.environ.get("IS_DOCKER_CONTAINER") == "1"


def is_snap_package(package: str) -> bool:
    """
    Checks if a specified package name is installed as a snap package
    by checking if it's running inside a SNAP environment.

    :param package: Name of the package to check.
    :type package: str
    :return: True if the specified package is installed as a snap package, otherwise False.
    :rtype: bool
    """

    if package:
        return True if os.getenv("SNAP") else False
    else:
        notify.error(INVALID_PACKAGE_ERROR.format(package=package))


def is_pypi_package(package: str) -> bool:
    """
    Checks if a specified package name is installed as a pypi package.

    :param package: Name of the package to check.
    :type package: str
    :return: True if the specified package is installed as a pypi package, otherwise False.
    :rtype: bool
    """

    if package:
        try:
            __import__(name=package)
            return True
        except ImportError:
            return False
    else:
        notify.error(INVALID_PACKAGE_ERROR.format(package=package))


def update_package(package: str, status: Optional[Status] = None):
    """
    Updates a specified pypi package.

    :param package: Name of the pypi package to update.
    :type package: str
    :param status: An optional `Status` object for displaying status messages.
    :type status: rich.status.Status
    """

    if package:
        try:
            if status:
                status.start()
                status.update("Downloading updates")
            subprocess.run(
                [
                    "pip",
                    "install",
                    "--upgrade",
                    package,
                ],
                check=True,
                stdout=subprocess.DEVNULL if status else None,
                stderr=subprocess.STDOUT if status else None,
            )
            notify.ok(f"DONE. The updates will be installed on next run.")
        except subprocess.CalledProcessError as called_process_error:
            notify.exception(
                title=f"An error occurred ({style.italic}while updating package{style.reset})",
                error=called_process_error,
            )
        except Exception as unexpected_error:
            notify.exception(unexpected_error)
    else:
        notify.error(INVALID_PACKAGE_ERROR.format(package=package))


# -------------------------------- END ----------------------------------------- #
